"""
functions and object for constants
"""
#!/usr/bin/python

# import official python packages
import logging
import os
import struct

# import custom python packages
from wahoomc.constants import RESOURCES_DIR
from wahoomc.constants import TOOLING_WIN_DIR
from wahoomc.constants import USER_CONFIG_DIR
from wahoomc.constants import USER_TOOLING_WIN_DIR
from wahoomc.file_directory_functions import read_json_file_generic

log = logging.getLogger('main-logger')


class TagWahooXmlNotFoundError(Exception):
    """Raised when the specified tag-wahoo xml file does not exist"""


class TagsToKeepNotFoundError(Exception):
    """Raised when the specified tags to keep .json file does not exist"""


def translate_tags_to_keep(name_tags=False, sys_platform='', use_repo=False):
    """
    translates the given tags to format of the operating system.
    """

    if sys_platform == "Windows":
        separator = ' ='
    else:
        separator = ', '

    tags_modif = []

    # read tags-to-keep .json from user-dir in favor of python installation
    # evaluate path first: user-dir in favor of PyPI installation
    if not use_repo:
        for path in get_absolute_dir_user_or_repo('', file='tags-to-keep.json'):
            if os.path.exists(path):
                break
    # force using file from repo - used in unittests for equal output
    else:
        path = get_absolute_dir_user_or_repo(
            '', file='tags-to-keep.json')[1]

    # read the tags from the evaluated path above
    tags_from_json = read_json_file_generic(path)

    if not tags_from_json:
        raise TagsToKeepNotFoundError

    if not name_tags:
        universal_tags = tags_from_json['TAGS_TO_KEEP_UNIVERSAL']
    else:
        universal_tags = tags_from_json['NAME_TAGS_TO_KEEP_UNIVERSAL']

    for tag, value in universal_tags.items():
        to_append = transl_tag_value(sys_platform, separator, tag, value)

        tags_modif.append(to_append)

    if sys_platform == "Windows":
        tags_modif = ' '.join(tags_modif)

    return tags_modif


def transl_tag_value(sys_platform, separator, tag, value):
    """
    translates one tag with value(s) to a "common" format
    """
    if isinstance(value, list):
        for iteration, sing_val in enumerate(value):
            if iteration == 0:
                to_append = f'{tag}={sing_val}'
            else:
                to_append = f'{to_append}{separator}{sing_val}'
    elif value:
        to_append = f'{tag}={value}'
    else:
        if sys_platform == "Windows":
            to_append = f'{tag}='
        else:
            to_append = tag

    return to_append


def get_tooling_win_path(path_in_tooling_win, in_user_dir=False):
    """
    return path to a tooling in the tooling_win directory and the given path
    OR from the user tooling_win directory
    """
    if in_user_dir:
        tooling_dir = USER_TOOLING_WIN_DIR
    else:
        tooling_dir = TOOLING_WIN_DIR

    # special for osmconvert: handle 32 and 64 bit here
    if path_in_tooling_win in ('osmconvert', 'osmconvert.exe'):
        if 8 * struct.calcsize("P") == 32:
            return os.path.join(tooling_dir, path_in_tooling_win)
        # 64 bit: replace with 64 in the end
        return os.path.join(tooling_dir, path_in_tooling_win.replace("osmconvert", "osmconvert64-0.8.8p"))

    # all other "toolings": concatenate with win tooling dir
    return os.path.join(tooling_dir, path_in_tooling_win)


def get_tag_wahoo_xml_path(tag_wahoo_xml):
    """
    return path to tag-wahoo xml file if the file exists
    - from the user directory "USER_WAHOO_MC/_config/tag_wahoo_adjusted/tag_wahoo_xml"
    - 2ndly from the PyPI installation: "RESOURCES_DIR/tag_wahoo_adjusted/tag_wahoo_xml"
    """

    for path in get_absolute_dir_user_or_repo("tag_wahoo_adjusted", tag_wahoo_xml):
        if os.path.exists(path):
            return path

    raise TagWahooXmlNotFoundError


def get_absolute_dir_user_or_repo(folder, file=''):
    """
    return the absolute path to the folder (and file) in this priorization
    1. user dir
    2. wahoomc package dir

    Priorization is important later on because user- should always be used in favor of repo-dir!
    """
    absolute_paths = []
    if file:
        absolute_paths.append(os.path.join(
            USER_CONFIG_DIR, folder, file))
        absolute_paths.append(os.path.join(
            RESOURCES_DIR, folder, file))
    else:
        absolute_paths.append(os.path.join(USER_CONFIG_DIR, folder))
        absolute_paths.append(os.path.join(
            RESOURCES_DIR, folder))

    return absolute_paths



"""
tests for the downloader file
"""
import os
# import sys
import unittest
from shapely.geometry import shape  # pylint: disable=import-error

# import custom python packages
from wahoomc.geofabrik import CountryGeofabrik, XYCombinationHasNoCountries, XYGeofabrik
from wahoomc.geofabrik import calc_bounding_box_tiles
from wahoomc.downloader import Downloader
from wahoomc import constants
from wahoomc.geofabrik_json import CountyIsNoGeofabrikCountry, GeofabrikJson


def calc_tiles_via_geofabrik_json(input_argument):
    """
    calculate tiles using downloaded geofabrik file
    the "new" way of doing
    """
    o_geofabrik = CountryGeofabrik(input_argument)
    tiles_via_geofabrik_json = o_geofabrik.get_tiles_of_wanted_map()

    return tiles_via_geofabrik_json


def calc_tiles_via_geofabrik_json_xy(input_argument):
    """
    calculate tiles using downloaded geofabrik file via x/y attributes as input
    the "new" way of doing
    """
    o_geofabrik = XYGeofabrik(input_argument)
    tiles_via_geofabrik_json = o_geofabrik.get_tiles_of_wanted_map()

    return tiles_via_geofabrik_json


class TestGeofabrik(unittest.TestCase):
    """
    tests for the downloader file
    """

    def setUp(self):
        self.max_days_old = 14

        if not os.path.isfile(constants.GEOFABRIK_PATH):
            o_downloader = Downloader(24, False)
            o_downloader.download_geofabrik_file()

    def test_tiles_via_geofabrik_malta(self):
        """
        Test the retrieval of tiles via geofabrik URL against hardcoded data
        """
        item_0 = {'x': 137, 'y': 100, 'left': 12.65625, 'top': 36.59788913307021,
                  'right': 14.0625, 'bottom': 35.4606699514953, 'countries': ['italy', 'malta'],
                  'urls': ['https://download.geofabrik.de/europe/italy-latest.osm.pbf',
                           'https://download.geofabrik.de/europe/malta-latest.osm.pbf']}
        item_1 = {'x': 138, 'y': 100, 'left': 14.0625, 'top': 36.59788913307021,
                  'right': 15.46875, 'bottom': 35.4606699514953, 'countries': ['italy', 'malta'],
                  'urls': ['https://download.geofabrik.de/europe/italy-latest.osm.pbf',
                           'https://download.geofabrik.de/europe/malta-latest.osm.pbf']}

        geofabrik_tiles = calc_tiles_via_geofabrik_json('malta')

        self.assertEqual(item_0, geofabrik_tiles[0])
        self.assertEqual(item_1, geofabrik_tiles[1])

    def test_tiles_via_geofabrik_malta_xy(self):
        """
        Test the retrieval of tiles via geofabrik URL against hardcoded data
        """
        item_0 = [{'x': 138, 'y': 100, 'left': 14.0625, 'top': 36.59788913307021, 'right': 15.46875, 'bottom': 35.4606699514953, 'countries': [
            'italy', 'malta'], 'urls': ['https://download.geofabrik.de/europe/italy-latest.osm.pbf', 'https://download.geofabrik.de/europe/malta-latest.osm.pbf']}]

        geofabrik_tiles = calc_tiles_via_geofabrik_json_xy('138/100')

        self.assertEqual(item_0, geofabrik_tiles)

    def test_tiles_via_geofabrik_xy(self):
        """
        Test the retrieval of tiles via geofabrik URL against hardcoded data
        """
        geofabrik_tiles_exp = [{
            "x": 133,
            "y": 88,
            "left": 7.03125,
            "top": 48.92249926375824,
            "right": 8.4375,
            "bottom": 47.98992166741417,
            "countries": [
                "france",
                "germany"
            ],
            'urls': ['https://download.geofabrik.de/europe/france-latest.osm.pbf',
                     'https://download.geofabrik.de/europe/germany-latest.osm.pbf']
        }]

        geofabrik_tiles = calc_tiles_via_geofabrik_json_xy('133/88')

        self.assertEqual(geofabrik_tiles_exp, geofabrik_tiles)

    def test_bbox_and_shape_malta(self):
        """
        Test several small steps in geofabrik processing for countries
        """
        input_country = 'malta'

        bbox_exp = {'top_x': 137, 'top_y': 100, 'bot_x': 138, 'bot_y': 100}

        bbox_tiles_exp = [{'x': 137, 'y': 100, 'tile_left': 12.65625, 'tile_top': 36.59788913307021, 'tile_right': 14.0625, 'tile_bottom': 35.4606699514953}, {
            'x': 138, 'y': 100, 'tile_left': 14.0625, 'tile_top': 36.59788913307021, 'tile_right': 15.46875, 'tile_bottom': 35.4606699514953}]

        tiles_exp = [{'x': 137, 'y': 100, 'left': 12.65625, 'top': 36.59788913307021, 'right': 14.0625, 'bottom': 35.4606699514953, 'countries': ['italy', 'malta'], 'urls': ['https://download.geofabrik.de/europe/italy-latest.osm.pbf', 'https://download.geofabrik.de/europe/malta-latest.osm.pbf']},
                     {'x': 138, 'y': 100, 'left': 14.0625, 'top': 36.59788913307021, 'right': 15.46875, 'bottom': 35.4606699514953, 'countries': ['italy', 'malta'], 'urls': ['https://download.geofabrik.de/europe/italy-latest.osm.pbf', 'https://download.geofabrik.de/europe/malta-latest.osm.pbf']}]

        wanted_region_exp = 'MULTIPOLYGON (((14 35.7517, 14 36.26791, 14.07957 36.33296, 14.39231 36.29082, 14.8561 35.98808, 14.84002 35.7246, 14.50306 35.51985, 14 35.7517)))'

        o_geofabrik = CountryGeofabrik('malta')
        o_geofabrik.get_tiles_of_wanted_map()

        o_geofabrik_json = GeofabrikJson()

        wanted_region = shape(
            o_geofabrik_json.get_geofabrik_geometry(input_country))
        self.assertEqual(wanted_region_exp, str(wanted_region))

        bbox = o_geofabrik.compose_bouding_box(wanted_region.bounds)
        self.assertEqual(bbox_exp, bbox)

        bbox_tiles = calc_bounding_box_tiles(
            bbox)
        self.assertEqual(bbox_tiles_exp, bbox_tiles)

        tiles_of_input = o_geofabrik.find_needed_countries(
            bbox_tiles, input_country, wanted_region)
        self.assertEqual(tiles_exp, tiles_of_input)

    def test_bbox_and_shape_xy(self):
        """
        Test several small steps in geofabrik processing for X/Y combination
        """
        bbox_tiles_exp = [{'x': 138, 'y': 100, 'tile_left': 14.0625,
                           'tile_top': 36.59788913307021, 'tile_right': 15.46875, 'tile_bottom': 35.4606699514953}]
        wanted_region_exp = 'POLYGON ((36.59788913307021 14.0625, 36.59788913307021 15.46875, 35.4606699514953 15.46875, 35.4606699514953 14.0625, 36.59788913307021 14.0625))'

        bbox = {'top_x': 138, 'bot_x': 138,
                'top_y': 100, 'bot_y': 100}

        bbox_tiles = calc_bounding_box_tiles(bbox)
        self.assertEqual(bbox_tiles_exp, bbox_tiles)

        o_geofabrik = XYGeofabrik('138/100')
        wanted_region_string = str(o_geofabrik.compose_shape(bbox_tiles))

        self.assertEqual(wanted_region_exp, wanted_region_string)

    def test_splitting_of_single_xy_coordinate(self):
        """
        use static json files in the repo to calculate relevant tiles
        """

        xy_tuple = XYGeofabrik.split_input_to_list("133/88")
        self.assertEqual(xy_tuple, [{"x": 133, "y": 88}])

        xy_tuple = XYGeofabrik.split_input_to_list("11/92")
        self.assertEqual(xy_tuple, [{"x": 11, "y": 92}])

        xy_tuple = XYGeofabrik.split_input_to_list("138/100")
        self.assertEqual(xy_tuple, [{"x": 138, "y": 100}])

    def test_splitting_of_multiple_xy_coordinate(self):
        """
        use static json files in the repo to calculate relevant tiles
        """

        xy_tuple = XYGeofabrik.split_input_to_list("133/88,138/100")
        expected_result = [{"x": 133, "y": 88}, {"x": 138, "y": 100}]

        self.assertEqual(xy_tuple, expected_result)

    def test_if_countrygeofabrik_raises_exception(self):
        """
        initialize CountryGeofabrik class and let the class check if the country is OK
        in other words: if the exception raised by translate_id_no_to_geofabrik
        is transported all up
        """
        CountryGeofabrik('germany')
        CountryGeofabrik('malta')
        CountryGeofabrik('malta,germany')

        with self.assertRaises(CountyIsNoGeofabrikCountry):
            CountryGeofabrik('germanyd')
        with self.assertRaises(CountyIsNoGeofabrikCountry):
            CountryGeofabrik('xy')
        with self.assertRaises(CountyIsNoGeofabrikCountry):
            CountryGeofabrik('maltad,germany')
        with self.assertRaises(CountyIsNoGeofabrikCountry):
            CountryGeofabrik('malta,tekke')

    def test_if_xy_geofabrik_raises_exception(self):
        """
        initialize XYGeofabrik class + check if xy coordinate has countries
        """
        o_geofabrik = XYGeofabrik('133/87')
        o_geofabrik.get_tiles_of_wanted_map()

        o_geofabrik = XYGeofabrik('100/138')
        o_geofabrik.get_tiles_of_wanted_map()

        o_geofabrik = XYGeofabrik('138/100')
        o_geofabrik.get_tiles_of_wanted_map()

        with self.assertRaises(XYCombinationHasNoCountries):
            o_geofabrik = XYGeofabrik('200/1')
            o_geofabrik.get_tiles_of_wanted_map()

    # def test_get_tile_via_xy_coordinate_error(self):
    #     """
    #     use static json files in the repo to calculate a not-existing tile.

    #     does not error out due to new Geofabrik Json processing. Nevertheless, the tile is not existing
    #     only +/- 180 -/+90: https://epsg.io/4326
    #     """

    #     o_geofabrik = XYGeofabrik([{"x": 200, "y": 1}])

    #     # tiles =

    #     with self.assertRaises(TileNotFoundError):
    #         o_geofabrik.get_tiles_of_wanted_map()


if __name__ == '__main__':
    unittest.main()



"""
constants, functions and object for file-system operations
"""
#!/usr/bin/python

# import official python packages
import json
import os
from os.path import isfile, join
import sys
import logging
import shutil

# import custom python packages

log = logging.getLogger('main-logger')


def move_content(src_folder_name, dst_path):
    """
    copy files from source directory of to destination directory
    delete source directory afterwards

    similar function to copy_or_move_files_and_folder but without user-request when overwriting
    and only support for directories
    """
    # build path to old folder on the same level as wahooMapsCreator
    par_dir = os.path.abspath(os.path.join(os.path.join(
        os.path.dirname(__file__), os.pardir), os.pardir))
    source_dir = os.path.join(par_dir, src_folder_name)

    if os.path.exists(source_dir):
        # copy & delete directory
        for item in os.listdir(source_dir):
            src = os.path.join(source_dir, item)
            dst = os.path.join(dst_path, item)
            # next, if destination directory exists
            if os.path.isdir(dst):
                continue

            if os.path.isdir(src):
                shutil.copytree(src, dst)
            else:
                shutil.copy2(src, dst)

        shutil.rmtree(source_dir)


def create_empty_directories(parent_dir, tiles_from_json, border_countries):
    """
    create empty directories for each tile and each country
    """
    for tile in tiles_from_json:
        outdir = os.path.join(parent_dir,
                              f'{tile["x"]}', f'{tile["y"]}')
        os.makedirs(outdir, exist_ok=True)

    for country in border_countries:
        outdir = os.path.join(parent_dir, country)
        os.makedirs(outdir, exist_ok=True)


def read_json_file_country_config(json_file_path):
    """
    read the country config (of last run) from the given json file
    """

    log.debug('-' * 80)
    log.debug('# Read country config json file')

    country_config = read_json_file_generic(json_file_path)
    if country_config == '':
        log.error('! Json file could not be opened.')
        sys.exit()

    log.debug(
        '+ Use country config file %s', json_file_path)
    log.debug('+ Read country config json file: OK')

    return country_config


def read_json_file_generic(json_file_path):
    """
    reads content of given .json file
    """
    try:
        with open(json_file_path, encoding="utf-8") as json_file:
            json_content = json.load(json_file)
            json_file.close()

        return json_content

    except FileNotFoundError:
        return {}


def write_json_file_generic(json_file_path, json_content):
    """
    writes content to .json file
    """
    # Serializing json
    json_content = json.dumps(json_content, indent=4)

    # Writing to file
    with open(json_file_path, "w", encoding="utf-8") as json_file:
        json_file.write(json_content)
        json_file.close()


def get_files_in_folder(folder):
    """
    return filenames of given folder without path as list
    """
    onlyfiles = [f for f in os.listdir(folder) if isfile(join(folder, f))]

    return onlyfiles


def delete_o5m_pbf_files_in_folder(folder):
    """
    delete .o5m and .pbf files of given folder
    """
    onlyfiles = [f for f in os.listdir(folder) if isfile(join(folder, f))]

    for file in onlyfiles:
        if file.endswith('.o5m') or file.endswith('.pbf'):
            try:
                os.remove(os.path.join(folder, file))
            except OSError:
                pass


def copy_or_move_files_and_folder(from_path, to_path, delete_from_dir=False):
    """
    copy content from source directory to destination directory
    optionally delete source directory afterwards

    similar function to move_content but with user-request when overwriting and support for files
    """
    # check if from path/file exists at all
    if os.path.exists(from_path):

        # given path is a directory
        if os.path.isdir(from_path):
            # first create the to-directory if not already there
            os.makedirs(to_path, exist_ok=True)

            for item in os.listdir(from_path):
                from_item = os.path.join(from_path, item)
                to_item = os.path.join(to_path, item)

                # copy directory
                if os.path.isdir(from_item):
                    copy_directory_w_user_input(from_item, to_item)

                # copy file
                else:
                    copy_file_w_user_input(from_item, to_item)

        # given path is a file
        else:
            copy_file_w_user_input(from_path, to_path)

        # directory
        if delete_from_dir:
            shutil.rmtree(from_path)


def copy_directory_w_user_input(from_item, to_item):
    """
    copy content from source directory to destination directory
    """
    if not os.path.isdir(to_item):
        shutil.copytree(from_item, to_item)
    else:
        val = input(f"{to_item} exists already. Overwrite? (y/n):")
        if val == 'y':
            shutil.copytree(from_item, to_item)
            log.info('! %s overwritten', to_item)
        else:
            log.debug('! %s not copied, exists already.', to_item)


def copy_file_w_user_input(from_item, to_item):
    """
    copy source file to destination file
    """
    if not os.path.isfile(to_item):
        shutil.copy2(from_item, to_item)
    else:
        val = input(f"{to_item} exists already. Overwrite? (y/n):")
        if val == 'y':
            shutil.copy2(from_item, to_item)
            log.info('! %s overwritten', to_item)
        else:
            log.debug('! %s not copied, exists already.', to_item)



"""
functions and object for processing input via CLI and GUI
"""
#!/usr/bin/python

# import official python packages
import argparse
import sys
from platform import uname

# for gui
import tkinter as tk
from tkinter import ttk

# import custom python packages
from wahoomc.geofabrik_json import GeofabrikJson
from wahoomc.geofabrik_json import CountyIsNoGeofabrikCountry
from wahoomc.geofabrik import CountryGeofabrik


def process_call_of_the_tool():
    """
    process CLI arguments
    """
    # input argument creation and processing
    desc = "Create up-to-date maps for your Wahoo ELEMNT and Wahoo ELEMNT BOLT"
    parser_top = argparse.ArgumentParser(description=desc)

    subparsers = parser_top.add_subparsers(title='Choose mode',
                                           description='choose the mode of using wahooMapsCreator. Either GUI or CLI.',
                                           help='sub-command help', dest='subparser_name')

    # create the parser for the "gui" command
    parser_gui = subparsers.add_parser(  # pylint: disable=unused-variable
        'gui', help='Start graphical user interface to select options')

    # create the parser for the "cli" command
    parser_cli = subparsers.add_parser(
        'cli', help='Run the tool via command line interface', formatter_class=argparse.RawTextHelpFormatter)

    # group: primary input parameters to create map for. One needs to be given
    primary_args = parser_cli.add_argument_group(
        title='Primary input', description='Generate maps for...')
    primary_args_excl = primary_args.add_mutually_exclusive_group(
        required=True)
    # country to create maps for
    primary_args_excl.add_argument(
        "-co", "--country", help="country to generate maps for.\nExample: -co malta, multiple countries separated by comma: -co malta,italy")
    # X/Y coordinates to create maps for
    primary_args_excl.add_argument(
        "-xy", "--xy_coordinates", help="x/y coordinates to generate maps for.\nExample: -xy 133/88, multiple xy coordinates separated by comma: -xy 133/88,134/89")

    # group: options for map generation
    options_args = parser_cli.add_argument_group(
        title='Options', description='Options for map generation')
    # Maximum age of source maps or land shape files before they are redownloaded
    options_args.add_argument('-md', '--maxdays', type=int, default=InputData().max_days_old,
                              help="maximum age of source maps and other files")
    # Do not calculate border countries of input country
    options_args.add_argument('-nbc', '--bordercountries', action='store_false',
                              help="do not process border countries of tiles involving more than one country")
    # calculate contour lines
    options_args.add_argument('-con', '--contour', action='store_true',
                              help="process contour lines (elevation data)")
    # use srtm1 for contour lines
    options_args.add_argument('-srtm1', '--use_srtm1', action='store_true',
                              help="use srtm1 as source for contour lines (elevation data)")
    # Force download of source maps and the land shape file
    # If False use Max_Days_Old to check for expired maps
    # If True force redownloading of maps and landshape
    options_args.add_argument('-fd', '--forcedownload', action='store_true',
                              help="force download of files")
    # Force (re)processing of source maps and the land shape file
    # If False only process files if not existing
    # If True force processing of files
    options_args.add_argument('-fp', '--forceprocessing', action='store_true',
                              help="force processing of files")
    # Save uncompressed maps for Cruiser if True
    options_args.add_argument('-c', '--cruiser', action='store_true',
                              help="save uncompressed maps for Cruiser")
    # specify the file with tags to keep in the output // file needs to be in wahoo_mc/resources/tag_wahoo_adjusted
    options_args.add_argument('-tag', '--tag_wahoo_xml', default=InputData().tag_wahoo_xml,
                              help="file with tags to keep in the output")
    # zip the country (and country-maps) folder
    options_args.add_argument('-z', '--zip', action='store_true',
                              help="zip the country (and country-maps) folder")
    options_args.add_argument('-v', '--verbose', action='store_true',
                              help="output debug logger messages")

    args = parser_top.parse_args()

    # process depending on GUI or CLI processing.
    # returns the input parameters in both cases
    if args.subparser_name == 'gui':
        # Prevents the initialisation of the graphical GUI on WSL.
        if 'microsoft' in uname().release:
            sys.exit("GUI can not be startet because no graphical interface is available. Start with 'python -m wahoo_mc cli -h' or 'python -m wahoo_mc -h' to see command line options.")

        o_input_data = GuiInput().start_gui()
        return o_input_data

    # cli processing
    o_input_data = InputData()
    o_input_data.country = args.country
    o_input_data.xy_coordinates = args.xy_coordinates
    o_input_data.max_days_old = args.maxdays

    o_input_data.process_border_countries = args.bordercountries
    o_input_data.contour = args.contour
    o_input_data.use_srtm1 = args.use_srtm1

    o_input_data.force_download = args.forcedownload
    o_input_data.force_processing = args.forceprocessing

    o_input_data.tag_wahoo_xml = args.tag_wahoo_xml
    o_input_data.save_cruiser = args.cruiser
    o_input_data.zip_folder = args.zip

    o_input_data.verbose = args.verbose

    return o_input_data


def cli_init():
    """
    Provides cli for initialization of user directory
    """

    parser = argparse.ArgumentParser(
        description='Copy config files to user directory')
    parser.add_argument('-v', '--verbose', action='store_true',
                        help="output debug logger messages")

    args = parser.parse_args()

    # cli processing
    o_input_data = InputData()
    o_input_data.verbose = args.verbose

    return o_input_data


def create_checkbox(self, default_value, description, row):
    """
    this is a reuse function for creating checkboxes.
    it
    - creates the checkbox in the given context "self"
    - returns the variable which can be accessed to get the value
      of the checkbox after clicking a button
    """
    bool_var = tk.BooleanVar()
    bool_var.set(default_value)

    # set checkbutton "anonymous". bool_var is later accessed to get the checkbox-value
    self.check_button = tk.Checkbutton(self, text=description,
                                       var=bool_var)
    self.check_button.grid(column=0, row=row, sticky=tk.W, padx=15, pady=5)

    return bool_var


def get_countries_of_continent_from_geofabrik(continent):
    """
    returns all countries of a continent to be selected in UI
    """
    countries = []
    for region, value in GeofabrikJson().geofabrik_overview.items():
        try:
            if value['parent'] == continent:
                countries.append(region)
        # regions/ continents do not have a parent
        except KeyError:
            pass

    return countries


class InputData():  # pylint: disable=too-many-instance-attributes,too-few-public-methods
    """
    object with all parameters to process maps and default values
    """

    def __init__(self):
        self.country = ""
        self.xy_coordinates = ""
        self.max_days_old = 14

        self.force_download = False
        self.force_processing = False
        self.process_border_countries = True
        self.contour = False
        self.use_srtm1 = False

        self.tag_wahoo_xml = "tag-wahoo-poi.xml"
        self.zip_folder = False
        self.save_cruiser = False

        self.verbose = False

    def is_required_input_given_or_exit(self, issue_message):
        """
        check, if the minimal required arguments is given:
        - country
        - x/y coordinates
        - file with tile coordinates
        If not, depending on the import parameter, the
        """
        if (self.country in ('None', '') and self.xy_coordinates in ('None', '')):
            if issue_message:
                sys.exit("Nothing to do. Start with -h or --help to see command line options."
                         "Or in the GUI select a country to create maps for.")
            else:
                sys.exit()
        elif self.country and self.xy_coordinates:
            sys.exit(
                "Country and X/Y coordinates are given. Only one of both is allowed!")
        elif self.country:
            # countries =
            try:
                CountryGeofabrik.split_input_to_list(self.country)
            except CountyIsNoGeofabrikCountry as exception:
                sys.exit(exception)

            # if we made it until here, sys.exit() was not called and therefore all countries OK ;-)
            return True
        else:
            return True


class GuiInput(tk.Tk):
    """
    This is the class to proces user-input via GUI
    """

    def __init__(self, *args, **kwargs):
        self.o_input_data = InputData()

        tk.Tk.__init__(self, *args, **kwargs)

    def start_gui(self):
        """
        start GUI
        """
        self.build_gui()

        # start GUI
        self.mainloop()

        self.o_input_data.is_required_input_given_or_exit(issue_message=True)
        return self.o_input_data

    def build_gui(self):
        """
        builds GUI consisting of several parts
        """
        # build GUI
        self.title("Wahoo map creator")
        self.configure(bg="white")

        # create notebook + tabs
        tab_control = ttk.Notebook(self, name="notebook")
        tab1 = tk.Frame(tab_control, name="tab1")
        tab2 = tk.Frame(tab_control, name="tab2")

        # add tabs to notebook & position
        tab_control.add(tab1, text="General settings")
        tab_control.add(tab2, text="Advanced settings")
        tab_control.pack(expand=1, fill="both")

        # content of 1. tab
        tab1.first = ComboboxesEntryField(
            tab1, self.o_input_data)
        tab1.first.pack(side=tk.TOP, fill=tk.X)

        tab1.third = CheckbuttonsTab1(
            tab1, self.o_input_data, controller=self)
        tab1.third.pack(side=tk.TOP, fill=tk.X)

        tab1.four = Buttons(tab1, controller=self)
        tab1.four.pack(side=tk.TOP, fill=tk.X)

        # content of 2. tab
        tab2.second = Text(tab2, self.o_input_data)
        tab2.second.pack(side=tk.TOP, fill=tk.X)

        tab2.first = CheckbuttonsTab2(
            tab2, self.o_input_data)
        tab2.first.pack(side=tk.TOP, fill=tk.X)

    def handle_create_map(self, event):  # pylint: disable=unused-argument
        """
        run when Button "Create" is pressed
        """
        # get tab1 and tab2 using "name" of the tkinker objects
        tab1 = self.children["notebook"].children["tab1"]
        tab2 = self.children["notebook"].children["tab2"]

        # you can get children of tkinker objects using ".winfo_children()"
        # other possibility for the above tab1 assignment:
        # tab1 = self.winfo_children()[0].winfo_children()[0]

        self.o_input_data.country = tab1.first.cb_country.get()
        self.o_input_data.max_days_old = int(tab1.first.input_maxdays.get())

        self.o_input_data.force_download = tab1.third.checkb_download.get()
        self.o_input_data.force_processing = tab1.third.checkb_processing_val.get()
        self.o_input_data.process_border_countries = tab1.third.checkb_border_countries_val.get()
        self.o_input_data.contour = tab1.third.checkb_contour_val.get()
        self.o_input_data.use_srtm1 = tab1.third.checkb_srtm1_val.get()

        self.o_input_data.save_cruiser = tab2.first.checkb_save_cruiser_val.get()
        self.o_input_data.zip_folder = tab2.first.checkb_zip_folder_val.get()
        self.o_input_data.verbose = tab2.first.checkb_verbose_val.get()

        # get text without \n in the end
        self.o_input_data.tag_wahoo_xml = tab2.second.input_tag_wahoo_xml.get()

        self.destroy()

    def switch_reload(self, event):  # pylint: disable=unused-argument
        """
        switch edit-mode of max-days field
        """
        tab1 = self.children["notebook"].children["tab1"]

        if tab1.first.en_max_days_old['state'] == tk.NORMAL:
            tab1.first.en_max_days_old.configure(state=tk.DISABLED)
        else:
            tab1.first.en_max_days_old.configure(state=tk.NORMAL)


class ComboboxesEntryField(tk.Frame):  # pylint: disable=too-many-instance-attributes
    """
    Comboboxes and Entry-Field for max days
    """

    def __init__(self, parent, oInputData):
        tk.Frame.__init__(self, parent)

        # Labels
        self.lab_top = tk.Label(
            self, text="Select continent and country to create a map")
        self.lab_continent = tk.Label(self, text="Select continent:")
        self.lab_country = tk.Label(self, text='Select country:')

        # Comboboxes
        self.cb_continent = ttk.Combobox(
            self, values=GeofabrikJson().geofabrik_regions, state="readonly")
        self.cb_continent.current(0)  # pre-select first entry in combobox
        self.cb_continent.bind("<<ComboboxSelected>>", self.callback_continent)

        self.cb_country = ttk.Combobox(
            self, state="readonly")

        # Positioning
        self.lab_top.grid(column=0, row=0, columnspan=2, padx=5, pady=10)

        self.lab_continent.grid(column=0, row=1, sticky=tk.E, padx=5, pady=2)
        self.cb_continent.grid(column=1, row=1, sticky=tk.W, padx=10, pady=2)

        self.lab_country.grid(column=0, row=2, sticky=tk.E, padx=5, pady=2)
        self.cb_country.grid(column=1, row=2, sticky=tk.W, padx=10, pady=2)

        self.lab_max_days_old = tk.Label(self, text='Max Old Days:')

        self.input_maxdays = tk.StringVar()
        self.input_maxdays.set(str(oInputData.max_days_old))

        self.en_max_days_old = tk.Entry(
            self, textvar=self.input_maxdays, width=5)

        self.lab_max_days_old.grid(
            column=0, row=3, sticky=tk.E, padx=5, pady=2)
        self.en_max_days_old.grid(column=1, row=3, sticky=tk.W, padx=10)

    def callback_continent(self, event):  # pylint: disable=unused-argument
        """
        set value-list of countries after changing continent
        """
        continent = self.cb_continent.get()
        # get countries for selected region and set for combobox
        self.cb_country["values"] = get_countries_of_continent_from_geofabrik(
            continent)
        self.cb_country.current(0)


class CheckbuttonsTab1(tk.Frame):
    """
    Checkbuttons for GUI - tab 2
    """

    def __init__(self, parent, oInputData, controller):
        tk.Frame.__init__(self, parent)
        self.controller = controller

        self.checkb_download = tk.BooleanVar()
        self.checkb_download.set(oInputData.force_download)

        self.chk_force_download = tk.Checkbutton(self, text="Force download",
                                                 var=self.checkb_download)
        self.chk_force_download.bind(
            "<Button-1>", self.controller.switch_reload)

        self.checkb_border_countries_val = create_checkbox(self, oInputData.process_border_countries,
                                                           "Process border countries", 0)
        self.checkb_contour_val = create_checkbox(self, oInputData.contour,
                                                  "process contour lines (elevation data)", 1)
        self.checkb_srtm1_val = create_checkbox(self, oInputData.use_srtm1,
                                                "use srtm1 as source for contour lines (elevation data)", 2)

        self.chk_force_download.grid(
            column=0, row=3, sticky=tk.W, padx=15, pady=5)
        self.checkb_processing_val = create_checkbox(self, oInputData.force_processing,
                                                     "Force processing", 4)


class Buttons(tk.Frame):
    """
    Buttons for GUI
    """

    def __init__(self, parent, controller):
        tk.Frame.__init__(self, parent)
        self.controller = controller
        self.btn_ok = tk.Button(self, text="Create map")
        self.btn_ok.bind("<Button-1>", self.controller.handle_create_map)

        self.btn_cancel = tk.Button(
            self, text="Exit", command=parent.master.master.destroy)

        self.btn_ok.pack(side=tk.LEFT, fill=tk.X, expand=True)
        self.btn_cancel.pack(side=tk.RIGHT, fill=tk.X, expand=True)


class Text(tk.Frame):
    """
    Text for GUI - 2. tab
    """

    def __init__(self, parent, oInputData):
        tk.Frame.__init__(self, parent)

        self.lab_tag_wahoo_xml = tk.Label(self, text='Tag wahoo XML file:')

        self.input_tag_wahoo_xml = tk.StringVar()
        self.input_tag_wahoo_xml.set(oInputData.tag_wahoo_xml)

        self.en_max_days_old = tk.Entry(
            self, textvar=self.input_tag_wahoo_xml, width=20)

        self.lab_tag_wahoo_xml.grid(
            column=0, row=1, sticky=tk.E, padx=5, pady=2)
        self.en_max_days_old.grid(column=1, row=1, sticky=tk.W, padx=10)


class CheckbuttonsTab2(tk.Frame):
    """
    Checkbuttons for GUI - 2. tab
    """

    def __init__(self, parent, oInputData):
        tk.Frame.__init__(self, parent)

        self.checkb_save_cruiser_val = create_checkbox(self, oInputData.save_cruiser,
                                                       "Save uncompressed maps for Cruiser", 2)
        self.checkb_zip_folder_val = create_checkbox(self, oInputData.zip_folder,
                                                     "Zip folder with generated files", 3)
        self.checkb_verbose_val = create_checkbox(self, oInputData.verbose,
                                                  "output debug logger messages", 4)



"""
tests for the constants geofabrik & geofabrik file
"""
import os
import unittest

import geojson  # pylint: disable=import-error

# import custom python packages
from wahoomc.downloader import Downloader
from wahoomc import constants
from wahoomc.geofabrik_json import GeofabrikJson
from wahoomc.geofabrik_json import CountyIsNoGeofabrikCountry


class TestConstantsGeofabrik(unittest.TestCase):
    """
    tests for constants in geofabrik context
    """

    def setUp(self):
        if not os.path.isfile(constants.GEOFABRIK_PATH):
            o_downloader = Downloader(24, False)
            o_downloader.download_geofabrik_file()

        self.o_geofabrik_json = GeofabrikJson()

        self.relevant_countries = []

        # calc relevant countries in constructor. these should be valid for the whole class
        with open(constants.GEOFABRIK_PATH, encoding='utf8') as file_handle:
            raw_json = geojson.load(file_handle)
        file_handle.close()

        for entry in raw_json.features:
            id_no = entry.properties['id']
            self.relevant_countries.append(id_no)

    def test_if_json_countries_exist_in_geofabrik(self):
        """
        go through all files in the wahoo_mc/resources/json directory
        - check if each country does also exist in geofabrik
        - some countries are skipped because they do not exist in geofabrik
        """
        for country in self.relevant_countries:
            child = self.o_geofabrik_json.translate_id_no_to_geofabrik(country)

            self.assertIn(child, {country.replace(
                '_', '-'), 'us/'+country.replace('_', '-')})

    def test_regions_of_geofabrik(self):
        """
        go through all files in the wahoo_mc/resources/json directory
        - check if each country does also exist in geofabrik
        - some countries are skipped because they do not exist in geofabrik
        """

        id_with_no_parent_geofabrik = []
        regions_geofabrik = []

        geofabrik_regions = ['africa', 'antarctica', 'asia', 'australia-oceania',
                             'central-america', 'europe', 'north-america', 'south-america', 'russia']

        # check against (new) raw representation of geofabrik json
        for feature in self.o_geofabrik_json.raw_json.features:
            try:
                feature.properties['parent']
            except KeyError:
                id_with_no_parent_geofabrik.append(feature.properties['id'])

        self.assertCountEqual(geofabrik_regions,
                              id_with_no_parent_geofabrik)

        # also check against (new) created dict
        for region in self.o_geofabrik_json.geofabrik_region_overview:
            regions_geofabrik.append(region)
        self.assertCountEqual(geofabrik_regions,
                              regions_geofabrik)

    def test_reading_geofabrik_parent(self):
        """
        go through all files in the wahoo_mc/resources/json directory
        - compare the built URL via existing coding with the url out of the geofabrik json file
        - some countries are skipped because they do not exist in geofabrik
        - in addition
            - skip countries if there is a geofabrik URL (because this URL is different to the built one)
            - skip countries which do not have a direct geofabrik json entry, mostly they are in another country/region
        """
        parent, child = self.o_geofabrik_json.get_geofabrik_parent_country(
            'germany')
        self.assertTrue(parent == 'europe' and child == 'germany')

        parent, child = self.o_geofabrik_json.get_geofabrik_parent_country(
            'baden-wuerttemberg')
        self.assertTrue(parent == 'germany' and child == 'baden-wuerttemberg')

        parent, child = self.o_geofabrik_json.get_geofabrik_parent_country(
            'malta')
        self.assertTrue(parent == 'europe' and child == 'malta')

        parent, child = self.o_geofabrik_json.get_geofabrik_parent_country(
            'asia')
        self.assertTrue(parent == '' and child == 'asia')

        with self.assertRaises(CountyIsNoGeofabrikCountry):
            parent, child = self.o_geofabrik_json.get_geofabrik_parent_country(
                'xy')

    def test_if_input_is_geofabrik_id(self):
        """
        check if the input is a id of the geofabrik file
        """
        self.assertTrue(
            self.o_geofabrik_json.is_input_a_geofabrik_id_no('germany'))

        self.assertTrue(
            self.o_geofabrik_json.is_input_a_geofabrik_id_no('baden-wuerttemberg'))

        self.assertTrue(
            self.o_geofabrik_json.is_input_a_geofabrik_id_no('asia'))
        self.assertFalse(
            self.o_geofabrik_json.is_input_a_geofabrik_id_no('xy'))

        self.assertFalse(self.o_geofabrik_json.is_input_a_geofabrik_id_no('*'))

    def test_tranlation_id_to_geofabrik_id(self):
        """
        get geofabrik id by country .json filename
        """
        transl_country = self.o_geofabrik_json.translate_id_no_to_geofabrik(
            'malta')
        self.assertEqual('malta', transl_country)

        transl_country = self.o_geofabrik_json.translate_id_no_to_geofabrik(
            'solomon_islands')
        self.assertEqual('solomon-islands', transl_country)

        transl_country = self.o_geofabrik_json.translate_id_no_to_geofabrik(
            'nebraska')
        self.assertEqual('us/nebraska', transl_country)


if __name__ == '__main__':
    unittest.main()



"""
constants, functions and object for file-system operations
"""
#!/usr/bin/python

# import official python packages
import os
import logging
import platform
import shutil
from pathlib import Path
import sys
import pkg_resources

# import custom python packages
from wahoomc.file_directory_functions import move_content, write_json_file_generic, \
    read_json_file_generic, delete_o5m_pbf_files_in_folder, copy_or_move_files_and_folder
from wahoomc.constants_functions import get_tooling_win_path, get_absolute_dir_user_or_repo
from wahoomc.downloader import get_latest_pypi_version

from wahoomc.constants import GEOFABRIK_PATH, USER_WAHOO_MC
from wahoomc.constants import USER_DL_DIR
from wahoomc.constants import USER_MAPS_DIR
from wahoomc.constants import USER_OUTPUT_DIR
from wahoomc.constants import USER_CONFIG_DIR
from wahoomc.constants import VERSION

log = logging.getLogger('main-logger')

config_file_path = os.path.join(USER_WAHOO_MC, ".config.json")


def initialize_work_directories():
    """
    Initialize work directories
    """
    os.makedirs(USER_WAHOO_MC, exist_ok=True)
    os.makedirs(USER_DL_DIR, exist_ok=True)
    os.makedirs(USER_MAPS_DIR, exist_ok=True)
    os.makedirs(USER_OUTPUT_DIR, exist_ok=True)
    os.makedirs(USER_CONFIG_DIR, exist_ok=True)


def move_old_content_into_new_dirs():
    """
    copy files from download- and output- directory of earlier version to the new folders
    delete directory from earlier versions afterwards

    having folder on the same level as the wahooMapsCreator was introduces in release v1.1.0 with PR #93.
    This coding is only valid/needed when using the cloned version or .zip version.
    If working with a installed version via PyPI, nothing will be done because folders to copy do not exist
    """
    move_content('wahooMapsCreator_download', USER_DL_DIR)
    move_content('wahooMapsCreator_output', USER_OUTPUT_DIR)


def adjustments_due_to_breaking_changes():
    """
    copy files from download- and output- directory of earlier version to the new folders
    """
    version_last_run = read_version_last_run()

    # file-names of filteres country files were uniformed in #153.
    # due to that old files are sometimes no longer accessed and files in the _tiles folder are deleted here.
    if version_last_run is None or \
            pkg_resources.parse_version(VERSION) <= pkg_resources.parse_version('2.0.2'):
        log.info(
            'Last run was with version %s, deleting files of %s directory due to breaking changes.', version_last_run, USER_OUTPUT_DIR)
        delete_o5m_pbf_files_in_folder(USER_OUTPUT_DIR)

    # file-names of downloaded .osm.pbf raw mapfiles was adjusted in #182 to focus on geofabrik naming
    # other existing files may therefor not be accessed anymore in the future and therefore deleted
    if version_last_run is None or \
            pkg_resources.parse_version(VERSION) < pkg_resources.parse_version('4.0.0a0'):
        log.info(
            'Last run was with version %s, deleting files of %s directory due to breaking changes.', version_last_run, USER_MAPS_DIR)
        delete_o5m_pbf_files_in_folder(USER_MAPS_DIR)
        log.info(
            'Last run was with version %s, deleting files of %s directory due to breaking changes.', version_last_run, USER_OUTPUT_DIR)
        delete_o5m_pbf_files_in_folder(USER_OUTPUT_DIR)


def check_installation_of_required_programs():
    """
    check if required programs are installed
    """
    text_to_docu = "\nPlease refer to the Quickstart Guide of wahooMapsCreator for instructions:\n- https://github.com/treee111/wahooMapsCreator/blob/develop/docs/QUICKSTART_ANACONDA.md \
                    \nor create an issue:\n- https://github.com/treee111/wahooMapsCreator/issues"

    if not is_program_installed("java"):
        sys.exit(
            f"Java is not installed. {text_to_docu}")

    if not os.path.isfile(GEOFABRIK_PATH):
        sys.exit('Geofabrik file is not downloaded. Please create an issue:\n- https://github.com/treee111/wahooMapsCreator/issues"')

    if platform.system() == "Windows":
        if not os.path.exists(get_tooling_win_path(
                os.path.join('Osmosis', 'bin', 'osmosis.bat'), in_user_dir=True)):
            sys.exit(
                f"Osmosis is not available. {text_to_docu}")

        if not os.path.exists(get_tooling_win_path('osmconvert.exe')):
            sys.exit(
                f"osmconvert is not available. {text_to_docu}")

        if not os.path.exists(get_tooling_win_path('osmfilter.exe', in_user_dir=True)):
            sys.exit(
                f"osmfilter is not available. {text_to_docu}")

        if not os.path.exists(get_tooling_win_path('7za.exe')):
            sys.exit(
                f"7za is not available. {text_to_docu}")

    else:
        if not is_program_installed("osmium"):
            sys.exit(
                f"osmium-tool is not installed. {text_to_docu}")

        if not is_program_installed("osmosis"):
            sys.exit(
                f"Osmosis is not installed. {text_to_docu}")

        if not is_map_writer_plugin_installed():
            sys.exit(
                f"mapsforge-map-writer plugin is not installed. {text_to_docu}")


def check_installation_of_programs_credentials_for_contour_lines():
    """
    check if additionals programs are installed
    """
    text_to_docu = "\nYou have choosen to process contour lines. That needs additional programs. \
                    \nPlease refer to the Quickstart Guide of wahooMapsCreator for instructions:\n- https://github.com/treee111/wahooMapsCreator/blob/develop/docs/QUICKSTART_ANACONDA.md#additional-programs-for-generating-contour-lines \
                    \nor create an issue:\n- https://github.com/treee111/wahooMapsCreator/issues"

    if not is_program_installed("phyghtmap"):
        sys.exit(
            f"phyghtmap is not installed. {text_to_docu}")

    username, password = read_earthexplorer_credentials()

    if not username or not password:
        username, password = ask_for_and_write_earthexplorer_credentials()


def is_program_installed(program):
    """
    check if a given program is installed
    """
    if shutil.which(program) is not None:
        return True

    return False


def is_map_writer_plugin_installed():
    """
    tests, if the mapwriter plugin is in the correct location

    Example filename for the map-writer-plugin
    mapsforge-map-writer-master-20210527.154736-408-jar-with-dependencies.jar
    downloaded on 01.10.2022: mapsforge-map-writer-0.18.0-jar-with-dependencies.jar
    """
    map_writer_path = os.path.join(
        str(Path.home()), '.openstreetmap', 'osmosis', 'plugins')

    # test if the file is there
    try:
        for file in os.listdir(map_writer_path):
            if "mapsforge-map-writer" in file:
                return True
    # if there is no file in the plugins directory
    except FileNotFoundError:
        pass

    return False


def write_config_file(config_to_write=''):
    """
    Write config file of wahoomc to root directory
    incorporate given content to existing file
    """
    # Data to be written
    default_config = {
        "version_last_run": VERSION
    }

    # if no config to write is given, write default config - normally at the end of main()
    if not config_to_write:
        config_to_write = default_config

    actual_content = read_json_file_generic(config_file_path)

    for key, value in config_to_write.items():
        # overwrite value or insert new item
        actual_content[key] = value

    # write changed content to disc
    write_json_file_generic(config_file_path, actual_content)


def read_version_last_run():
    """
    Read the version of wahoomc's last run
    by reading json and access version attribute, if not set, give None
    """
    try:
        version_last_run = read_json_file_generic(config_file_path)[
            "version_last_run"]
    except KeyError:
        version_last_run = None

    return version_last_run


def ask_for_and_write_earthexplorer_credentials():
    """
    Ask user for credentials for https://ers.cr.usgs.gov and save in the config file
    """
    log.warning(
        'No saved credentials found for https://ers.cr.usgs.gov. Please register and enter your credentials.')
    username = input('https://ers.cr.usgs.gov username:')
    password = input('https://ers.cr.usgs.gov password:')

    credentials_to_write = {
        'earthexplorer-user': username, 'earthexplorer-password': password}

    write_config_file(credentials_to_write)

    return username, password


def read_earthexplorer_credentials():
    """
    Read the version of wahoomc's last run
    by reading json and access version attribute, if not set, give None
    """
    try:
        username = read_json_file_generic(config_file_path)[
            "earthexplorer-user"]
        password = read_json_file_generic(config_file_path)[
            "earthexplorer-password"]
    except KeyError:
        username = None
        password = None

    return username, password


def copy_jsons_from_repo_to_user(folder, file=''):
    """
    copies files from wahoomc repo/package to the user-directory
    """
    absolute_paths = get_absolute_dir_user_or_repo(folder, file)

    log.debug('# Copy "%s" files from repo to directory if not existing: %s',
              folder, absolute_paths[0])

    # copy files of wahoomc package directory to user directory
    copy_or_move_files_and_folder(
        absolute_paths[1], absolute_paths[0], delete_from_dir=False)

    log.info('# Copy "%s" files from wahoomc installation to directory if not existing: %s : OK',
             folder, absolute_paths[0])


def check_installed_version_against_latest_pypi():
    """
    get latest wahoomc version available on PyPI and compare with locally installed version
    """
    # get latest wahoomc version available on PyPI
    latest_version = get_latest_pypi_version()

    # compare installed version against latest and issue a info if a new version is available
    if latest_version \
            and pkg_resources.parse_version(VERSION) < pkg_resources.parse_version(latest_version):
        log.info('\n\nUpdate available! \
                \nA new version of wahoomc is available: "%s". You have installed version "%s". \
                \nUpgrade wahoomc with "pip install wahoomc --upgrade". \
                \nRelease notes are here: https://github.com/treee111/wahooMapsCreator/releases/latest. \
                \n',
                 latest_version, VERSION)



"""
tests for the osm maps file
"""
import os
# import sys
import unittest

# import custom python packages
from wahoomc.osm_maps_functions import CountryOsmData, XYOsmData
from wahoomc.osm_maps_functions import OsmMaps
# from wahoomc.osm_maps_functions import TileNotFoundError
from wahoomc.input import InputData
from wahoomc import file_directory_functions as fd_fct
from wahoomc import constants


class TestOsmMapsCalculation(unittest.TestCase):
    """
    tests for the OSM maps file
    """

    # def setUp(self):

    def test_calc_border_countries_input_country(self):
        """
        Test initialized border countries
        - of malta
        - of germany
        """

        # malta
        self.process_and_check_border_countries(
            'malta', True, {'italy': {}, 'malta': {}}, 'country')

        # germany
        expected_result = {'czech-republic': {}, 'germany': {}, 'austria': {}, 'liechtenstein': {},
                           'switzerland': {}, 'italy': {}, 'netherlands': {}, 'belgium': {},
                           'luxembourg': {}, 'france': {}, 'poland': {}, 'denmark': {}, 'sweden': {}}
        self.process_and_check_border_countries(
            'germany', True, expected_result, 'country')

        # germany,malta
        expected_result = {'czech-republic': {}, 'germany': {}, 'austria': {}, 'liechtenstein': {},
                           'switzerland': {}, 'italy': {}, 'netherlands': {}, 'belgium': {},
                           'luxembourg': {}, 'france': {}, 'poland': {}, 'denmark': {}, 'sweden': {}, 'malta': {}}
        self.process_and_check_border_countries(
            'germany,malta', True, expected_result, 'country')

    def test_calc_border_countries_input_xy_coordinates_1tile(self):
        """
        Test initialized border countries
        - of a file with 1 tile and two countries
        """

        # one tile - france and germany
        expected_result = {'france': {}, 'germany': {}}
        self.process_and_check_border_countries(
            "133/88", True, expected_result, 'xy_coordinate')

    def test_calc_border_countries_input_xy_coordinates_2tiles(self):
        """
        Test initialized border countries
        - of a file with 2 tiles and one country
        """

        # two tiles - germany
        expected_result = {'germany': {}}
        self.process_and_check_border_countries(
            "134/87,134/88", True, expected_result, 'xy_coordinate')

    def test_calc_without_border_countries_input_country(self):
        """
        Test initialized countries without border countries
        - of germany
        - of china
        - of one tile
        """

        # germany
        self.process_and_check_border_countries(
            'germany', False, {'germany': {}}, 'country')

        # china
        self.process_and_check_border_countries(
            'china', False, {'china': {}}, 'country')

        # malta,liechtenstein
        self.process_and_check_border_countries(
            'malta,liechtenstein', False, {'malta': {}, 'liechtenstein': {}}, 'country')

        # malta,tunisia
        self.process_and_check_border_countries(
            'malta,tunisia', False, {'malta': {}, 'tunisia': {}}, 'country')

    def test_calc_without_border_countries_xy_coordinates_1tile(self):
        """
        Test initialized countries without border countries
        - of one tile
        """

        # one tile - france and germany
        expected_result = {'france': {}, 'germany': {}}
        self.process_and_check_border_countries(
            "133/88", False, expected_result, 'xy_coordinate')

    def test_calc_without_border_countries_xy_coordinates_2tiles(self):
        """
        Test initialized countries without border countries
        - of two tiles
        """

        # two tiles - germany
        expected_result = {'germany': {}}
        self.process_and_check_border_countries(
            "134/87,134/88", False, expected_result, 'xy_coordinate')

    def process_and_check_border_countries(self, inp_val, calc_border_c, exp_result, inp_mode):
        """
        helper method to process a country or json file and check the calculated border countries
        """

        o_input_data = InputData()
        o_input_data.process_border_countries = calc_border_c

        if inp_mode == 'country':
            o_input_data.country = inp_val
            o_osm_data = CountryOsmData(o_input_data)
        elif inp_mode == 'xy_coordinate':
            o_input_data.xy_coordinates = inp_val
            o_osm_data = XYOsmData(o_input_data)

        o_osm_data.process_input_of_the_tool()

        result = o_osm_data.border_countries

        # delete the path to the file, here, only the correct border countries are checked
        for res in result:
            result[res] = {}

        self.assertEqual(result, exp_result)


class TestOSMMapsInput(unittest.TestCase):
    """
    tests for input of OsmData
    """

    def test_input_country_malta(self):
        """
        Test "malta" as input to the wahooMapsCreator
        check, if the given input-parameter is saved to the OsmMaps instance
        """
        o_osm_data = self.get_osm_data_instance('malta')

        self.assertEqual(o_osm_data.country_name, 'malta')

    def test_folder_name_many_countries(self):
        """
        Test a very long list of countries as input to the wahooMapsCreator
        check, if the list of countries is cutted down to 100 chars and that is saved to the OsmMaps instance
        """
        o_osm_data = self.get_osm_data_instance('albania,alps,andorra,austria,azores,belarus,belgium,bosnia-herzegovina,britain-and-ireland,bulgaria,croatia,cyprus,czech-republic,dach,denmark,estonia,faroe-islands,finland,france,georgia,germany,great-britain,greece,guernsey-jersey,hungary,iceland,ireland-and-northern-ireland,isle-of-man,italy,kosovo,latvia,liechtenstein,lithuania,luxembourg,macedonia,malta,moldova,monaco,montenegro,netherlands,norway,poland,portugal,romania,serbia,slovakia,slovenia,spain,sweden,switzerland,turkey,ukraine')

        o_osm_maps = OsmMaps(o_osm_data)
        folder_name = o_osm_maps.calculate_folder_name('.map.lzma')
        folder_name_maps = o_osm_maps.calculate_folder_name('.map')

        self.assertEqual(folder_name, 'albania_alps_andorra_austria_azores_belarus_belgiu')
        self.assertEqual(folder_name_maps, 'albania_alps_andorra_austria_azores_belarus_belgiu-maps')

    def test_encoding_open_sea_osm(self):
        """
        use static json files in the repo to calculate relevant tile
        """

        with open(os.path.join(constants.RESOURCES_DIR, 'sea.osm')) as sea_file:  # pylint: disable=unspecified-encoding
            sea_data_no_encoding = sea_file.read()
        with open(os.path.join(constants.RESOURCES_DIR, 'sea.osm'), encoding="utf-8") as sea_file:
            sea_data_utf8 = sea_file.read()

        self.assertEqual(sea_data_no_encoding, sea_data_utf8)

    def get_osm_data_instance(self, country_input):
        """
        takes given input creates OsmData instance
        """
        o_input_data = InputData()
        o_input_data.country = country_input

        o_osm_data = CountryOsmData(o_input_data)
        o_osm_data.process_input_of_the_tool()

        return o_osm_data

class TestConfigFile(unittest.TestCase):
    """
    tests for the config .json file in the "wahooMapsCreatorData/_tiles/{country}" directory
    """

    def test_version_and_tags_of_country_config_file(self):
        """
        tests, if the return value of version is OK and if the tags are the same
        """

        o_input_data = InputData()
        o_input_data.country = 'malta'
        # prevent from downloading land_polygons each time
        o_input_data.max_days_old = 1000

        o_osm_data = CountryOsmData(o_input_data)
        o_osm_data.process_input_of_the_tool()

        o_downloader = o_osm_data.get_downloader()

        # download files marked for download to fill up map_file per country to write to config
        o_downloader.download_files_if_needed()

        o_osm_maps = OsmMaps(o_osm_data)

        o_osm_maps.write_country_config_file(o_input_data.country)

        self.assertTrue(
            o_osm_maps.tags_are_identical_to_last_run(o_input_data.country))

        country_config = fd_fct.read_json_file_country_config(os.path.join(
            constants.USER_OUTPUT_DIR, o_input_data.country, ".config.json"))

        self.assertEqual(constants.VERSION, country_config["version_last_run"])


if __name__ == '__main__':
    unittest.main()



"""
executable file to create up-to-date map-files for the Wahoo ELEMNT and Wahoo ELEMNT BOLT
"""
#!/usr/bin/python

# import official python packages
import logging

# import custom python packages
from wahoomc.input import process_call_of_the_tool, cli_init
from wahoomc.setup_functions import initialize_work_directories, \
    check_installation_of_required_programs, write_config_file, \
    adjustments_due_to_breaking_changes, copy_jsons_from_repo_to_user, \
    check_installed_version_against_latest_pypi, check_installation_of_programs_credentials_for_contour_lines
from wahoomc.downloader import download_tooling

from wahoomc.osm_maps_functions import OsmMaps
from wahoomc.osm_maps_functions import CountryOsmData, XYOsmData

# logging used in the terminal output:
# # means top-level command
# ! means error
# + means additional comment in a working-unit


def run(run_level):
    """
    main program run
    """
    # create logger
    logging.basicConfig(format='%(levelname)s:%(message)s',
                        level=logging.INFO)

    # initializing work directories needs to be the first call,
    # because other setup stuff relies on that (breaking changes)
    check_installed_version_against_latest_pypi()
    initialize_work_directories()
    adjustments_due_to_breaking_changes()
    download_tooling()
    check_installation_of_required_programs()

    if run_level == 'init':
        o_input_data = cli_init()
    else:
        # handle GUI and CLI processing via one function and different cli-calls
        o_input_data = process_call_of_the_tool()

    if o_input_data.verbose:
        logging.getLogger().setLevel(logging.DEBUG)

    if run_level == 'init':
        copy_jsons_from_repo_to_user('tag_wahoo_adjusted')
        copy_jsons_from_repo_to_user('.', 'tags-to-keep.json')
    else:
        # Is there something to do?
        o_input_data.is_required_input_given_or_exit(issue_message=True)

        if o_input_data.contour:
            check_installation_of_programs_credentials_for_contour_lines()

        if o_input_data.country:
            o_osm_data = CountryOsmData(o_input_data)
        elif o_input_data.xy_coordinates:
            o_osm_data = XYOsmData(o_input_data)

        # Check for not existing or expired files. Mark for download, if dl is needed
        o_osm_data.process_input_of_the_tool()
        o_downloader = o_osm_data.get_downloader()

        # Download files marked for download
        o_downloader.download_files_if_needed()

        o_osm_maps = OsmMaps(o_osm_data)

        # Filter tags from country osm.pbf files'
        o_osm_maps.filter_tags_from_country_osm_pbf_files()

        # Generate land
        o_osm_maps.generate_land()

        # Generate sea
        o_osm_maps.generate_sea()

        # Generate elevation
        if o_input_data.contour:
            o_osm_maps.generate_elevation(o_input_data.use_srtm1)

        # Split filtered country files to tiles
        o_osm_maps.split_filtered_country_files_to_tiles()

        # Merge splitted tiles with land an sea
        o_osm_maps.merge_splitted_tiles_with_land_and_sea(
            o_input_data.process_border_countries)

        # Creating .map files
        o_osm_maps.create_map_files(o_input_data.save_cruiser,
                                    o_input_data.tag_wahoo_xml)

        # Zip .map.lzma files
        o_osm_maps.make_and_zip_files('.map.lzma', o_input_data.zip_folder)

        # Make Cruiser map files zip file
        if o_input_data.save_cruiser is True:
            o_osm_maps.make_and_zip_files('.map', o_input_data.zip_folder)

    # run was successful --> write config file
    write_config_file()



"""
tests the generated files of a full-run of the tool against fixed files
"""

import filecmp
import os
from os import walk
import platform
import shutil
import unittest
import subprocess


# import custom python packages
from wahoomc import constants
from wahoomc.file_directory_functions import get_files_in_folder

dirname_of_file = os.path.dirname(__file__)
unittest_files_root = os.path.join(constants.USER_WAHOO_MC, '_unittest')
unittest_files_parking = os.path.join(unittest_files_root, 'parking_lot')


def copy_static_maps_input_file(mode, country, given_osm_pbf=''):
    """
    copy given file to download-directory
    """
    static_file_path = os.path.join(
        dirname_of_file, 'resources', given_osm_pbf)
    prod_path = os.path.join(
        constants.USER_MAPS_DIR, country + '-latest.osm.pbf')
    parking_path = os.path.join(
        unittest_files_parking, country + '-latest.osm.pbf')

    # these modes need the static filename to copy
    if mode in [0, 1] and given_osm_pbf == '':
        raise SystemError

    move_file_dir(mode, static_file_path, prod_path, parking_path)


def copy_static_land_polygon_input_folder(mode):
    """
    copy given folder to download-directory
    - mode 0: parking
    - mode 1: normal
    - mode 2: restore
    """
    static_file_path = os.path.join(
        unittest_files_root, 'land-polygons-split-4326_2021-10-31')
    prod_path = os.path.dirname(constants.LAND_POLYGONS_PATH)
    parking_path = os.path.join(
        unittest_files_parking, 'land-polygons-split-4326')

    move_file_dir(mode, static_file_path, prod_path, parking_path)


def copy_static_geofabrik_file(mode):
    """
    copy given file to download-directory
    - mode 0: parking
    - mode 1: normal
    - mode 2: restore
    """
    static_file_path = os.path.join(
        dirname_of_file, 'resources', 'geofabrik-2023-02-26.json')
    prod_path = os.path.join(constants.GEOFABRIK_PATH)
    parking_path = os.path.join(
        unittest_files_parking, 'geofabrik.json')

    move_file_dir(mode, static_file_path, prod_path, parking_path)


def move_file_dir(mode, static_file_path, prod_path, parking_path):
    """
    this function does the real file movements
    doing mode 0 and 1 in a while loop. Makes mostly sense together
    """
    while True:
        copy_from_path, copy_to_path = eval_from_to_paths(
            mode, static_file_path, prod_path, parking_path)

        # delete file or dir
        if os.path.isfile(copy_to_path):
            os.remove(copy_to_path)
        elif os.path.isdir(copy_to_path):
            # delete directory if exists. copytree fails if dir exists already
            shutil.rmtree(copy_to_path)
        else:  # not existing
            pass

        # copy from- to to- path file/dir
        try:
            if os.path.isfile(copy_from_path):
                # copy file
                shutil.copy2(copy_from_path, copy_to_path)
            elif os.path.isdir(copy_from_path):
                # copy folder
                shutil.copytree(copy_from_path, copy_to_path)
            else:  # not existing
                pass
        except FileNotFoundError:
            pass

        # either go another round or go out
        if mode in (1, 2):
            break
        if mode == 0:
            mode = 1


def eval_from_to_paths(mode, static_file_path, prod_path, parking_path):
    """
    evaluate from and to path for copying by given mode
    """
    if mode == 0:  # parking
        copy_from_path = prod_path
        copy_to_path = parking_path
    elif mode == 1:  # normal
        copy_from_path = static_file_path
        copy_to_path = prod_path
    elif mode == 2:  # restore
        copy_from_path = parking_path
        copy_to_path = prod_path
    else:
        raise SystemError

    return copy_from_path, copy_to_path


class TestGeneratedFiles(unittest.TestCase):
    """
    tests for the OSM maps file

    Test output of countries without border countries
    - Using given input .osm.pbf file in this repo
    - Compare content of directory tests/resources/mac /windows against directory output/

    Files that can be possibly checked (in order of creation):
    - merged.osm.pbf
    - 100.map
    - 100.map.lzma
    """

    def setUp(self):
        os.makedirs(unittest_files_parking, exist_ok=True)

        # copy actual productive files to parking lot
        # and copy static files as productive ones to have equal results each run
        copy_static_land_polygon_input_folder(mode=0)
        copy_static_geofabrik_file(mode=0)
        copy_static_maps_input_file(
            mode=0, country='malta', given_osm_pbf='malta-latest_2021-10-31.osm.pbf')
        copy_static_maps_input_file(
            mode=0, country='liechtenstein', given_osm_pbf='liechtenstein-latest_2021-10-31.osm.pbf')

    def tearDown(self):
        # copy files from parking lot back as productive
        copy_static_land_polygon_input_folder(mode=2)
        copy_static_geofabrik_file(mode=2)
        copy_static_maps_input_file(mode=2, country='malta')
        copy_static_maps_input_file(mode=2, country='liechtenstein')

    def test_calc_output_malta_and_compare(self):
        """
        Test output of countries without border countries
        - of malta
        """
        # run tool for country
        self.run_wahoomapscreator_cli('malta')

        # compare generated given merged.osm.pbf-file with reference-file
        self.compare_dir_sub_test_resource_and_output('138', 'malta')
        self.compare_dir_sub_test_resource_and_output('137', 'malta')

    def test_calc_output_liechtenstein_and_compare(self):
        """
        Test output of countries without border countries
        - of liechtenstein
        """
        # run tool for country
        self.run_wahoomapscreator_cli('liechtenstein')

        # compare generated given merged.osm.pbf-file with reference-file
        self.compare_dir_sub_test_resource_and_output('134', 'liechtenstein')

    def run_wahoomapscreator_cli(self, country):
        """
        runs wahooMapsCreator for given country and static osm.pbf file via CLI
        - without border countries
        """
        # run processing of input-country via CLI
        result = os.system(
            f'python -m wahoomc cli -co {country} -tag tag-wahoo.xml -fp -c -md 9999 -nbc')

        # check if run was successful
        self.assertEqual(result, 0)

    def test_malta_compare(self):
        """
        compare output of malta
        - of malta
        """
        # compare generated X/Y files with reference-file
        self.compare_dir_sub_test_resource_and_output('138', 'malta')
        self.compare_dir_sub_test_resource_and_output('137', 'malta')

    def test_liechtenstein_compare(self):
        """
        compare output of malta
        - of liechtenstein
        """
        # compare generated X/Y files with reference-file
        self.compare_dir_sub_test_resource_and_output('134', 'liechtenstein')

    def compare_dir_sub_test_resource_and_output(self, dir_to_compare, country):
        """
        compare files from given directory and subdirectories for equalness.
        check the files which are in test/resources folder
        """

        if platform.system() == "Windows":
            path_to_dir = os.path.join(
                dirname_of_file, 'resources', 'windows')
        else:
            path_to_dir = os.path.join(
                dirname_of_file, 'resources', 'macos')

        # check files in subdir of given dir (i.e. the X/Y directory)
        for (dirpath, dirnames, filenames) in walk(os.path.join(path_to_dir, dir_to_compare)):  # pylint: disable=unused-variable
            for directory in dirnames:
                for (dirpath_2, dirnames_2, filenames_2) in walk(os.path.join(  # pylint: disable=unused-variable
                        dirpath, directory)):
                    for file in filenames_2:
                        if file == '.DS_Store':
                            continue
                        given_output_file = os.path.join(
                            dirpath_2, file)
                        calculated_output_file = os.path.join(
                            constants.USER_OUTPUT_DIR, dir_to_compare, directory, file)

                        # are these two files equal?
                        self.compare_two_map_files(
                            given_output_file, calculated_output_file)

        # check files in given dir - {country} folder. filtered_* files
        for file in get_files_in_folder(os.path.join(path_to_dir, country)):
            if file == '.DS_Store':
                continue
            given_output_file = os.path.join(
                path_to_dir, country, file)
            calculated_output_file = os.path.join(
                constants.USER_OUTPUT_DIR, country, file)

            # are these two files equal?
            self.compare_two_map_files(
                given_output_file, calculated_output_file)

    def compare_two_map_files(self, given_file, calculated_file):
        """
        compare two given (map) files for equalness.
        macOS / Linux:
        - classic map files are compared using CLI command "osmium diff",
        - the others are compared using "filecmp.cmp"
        Windows:
        - compare all files using "filecmp.cmp".
          osmosis and osmconvert do not offer a possibility to compare with returning a errorcode
        """

        no_osmosis_file_extensions = ['shx', 'shp', 'prj']

        # some file extensions can not be comapared using osmium
        if given_file.split('.')[-1] in no_osmosis_file_extensions:
            # platform.system() == "Windows":
            self.assertTrue(filecmp.cmp(given_file, calculated_file,
                                        shallow=False), f'not equal: {calculated_file}. Using filecmp.cmp.')
        # compare map files using osmium
        else:
            cmd = ['osmium', 'diff', '-q',
                   given_file, calculated_file]
            result = subprocess.run(cmd, check=False)

            self.assertEqual(
                0, result.returncode, f'not equal: {calculated_file}. Using osmium diff.')


if __name__ == '__main__':
    unittest.main()



"""
tests for the python file
"""
import os
import unittest

# import custom python packages


class TestCli(unittest.TestCase):
    """
    tests for the CLI of the python file
    """

    def test_top_parser_help(self):
        """
        tests, if help of top parser can be called
        """

        result = os.system("python -m wahoomc -h")

        self.assertEqual(result, 0)

    def test_cli_help(self):
        """
        tests, if CLI help can be called
        """

        result = os.system("python -m wahoomc cli -h")

        self.assertEqual(result, 0)

    def test_gui_help(self):
        """
        tests, if GUI help can be called
        """

        result = os.system("python -m wahoomc gui -h")

        self.assertEqual(result, 0)


if __name__ == '__main__':
    unittest.main()



"""
constants
"""
#!/usr/bin/python

import os
from pathlib import Path

# User
USER_DIR = str(Path.home())
USER_WAHOO_MC = os.path.join(str(Path.home()), 'wahooMapsCreatorData')
USER_DL_DIR = os.path.join(USER_WAHOO_MC, '_download')
USER_MAPS_DIR = os.path.join(USER_DL_DIR, 'maps')
LAND_POLYGONS_PATH = os.path.join(
    USER_DL_DIR, 'land-polygons-split-4326', 'land_polygons.shp')
GEOFABRIK_PATH = os.path.join(USER_DL_DIR, 'geofabrik.json')
USER_OUTPUT_DIR = os.path.join(USER_WAHOO_MC, '_tiles')
USER_CONFIG_DIR = os.path.join(USER_WAHOO_MC, '_config')
USER_TOOLING_WIN_DIR = os.path.join(USER_DL_DIR, 'tooling_win')
OSMOSIS_WIN_FILE_PATH = os.path.join(
    USER_TOOLING_WIN_DIR, 'Osmosis', 'bin', 'osmosis.bat')

# Python Package - wahooMapsCreator directory
WAHOO_MC_DIR = os.path.dirname(__file__)
RESOURCES_DIR = os.path.join(WAHOO_MC_DIR, 'resources')
TOOLING_WIN_DIR = os.path.join(WAHOO_MC_DIR, 'tooling_win')
# location of repo / python installation - not used
ROOT_DIR = os.path.abspath(os.path.join(os.path.dirname(__file__), os.pardir))
VERSION = '4.1.1a0'


block_download = ['dach', 'alps', 'britain-and-ireland', 'south-africa-and-lesotho',
                  'us-midwest', 'us-northeast', 'us-pacific', 'us-south', 'us-west']

# Special_regions like (former) colonies where the map of the wanted region is not present in the map of the parent country.
# example Guadeloupe, it's Geofabrik parent country is France but Guadeloupe is not located within the region covered by the map of France.
special_regions = ['guadeloupe', 'guyane', 'martinique', 'mayotte', 'reunion']



"""
functions and object for checking for up-to-date & downloading land_polygons file and OSM maps
"""
#!/usr/bin/python

# import official python packages
import glob
import os
import os.path
import sys
import time
import logging
import platform
import zipfile
import requests

# import custom python packages
from wahoomc.constants_functions import get_tooling_win_path
from wahoomc.geofabrik_json import GeofabrikJson

from wahoomc.constants import USER_DL_DIR
from wahoomc.constants import USER_MAPS_DIR
from wahoomc.constants import LAND_POLYGONS_PATH
from wahoomc.constants import GEOFABRIK_PATH
from wahoomc.constants import OSMOSIS_WIN_FILE_PATH
from wahoomc.constants import USER_TOOLING_WIN_DIR
from wahoomc.constants import USER_DIR

log = logging.getLogger('main-logger')


def older_than_x_days(file_creation_timestamp, max_days_old):
    """
    check if given timestamp is older that given days
    """
    to_old_timestamp = time.time() - 60 * 60 * 24 * max_days_old

    return bool(file_creation_timestamp < to_old_timestamp)


def download_file(target_filepath, url, target_dir=""):
    """
    download given file and eventually unzip it
    """
    logging_filename = target_filepath.split(os.sep)[-1]
    log.info('-' * 80)
    log.info('# Downloading %s file', logging_filename)
    if url.split('.')[-1] == 'zip':
        # build target-filepath based on last element of URL
        last_part = url.rsplit('/', 1)[-1]
        dl_file_path = os.path.join(USER_DL_DIR, last_part)
        # download URL to file
        download_url_to_file(url, dl_file_path)

        # if a target directory is given --> extract into that folder
        if target_dir:
            target_path = target_dir
        else:
            target_path = USER_DL_DIR

        # unpack it
        unzip(dl_file_path, target_path)

        os.remove(dl_file_path)
    else:
        # no zipping --> directly download to given target filepath
        download_url_to_file(url, target_filepath)
    # Check if file exists (if target file exists)
    if not os.path.isfile(target_filepath):
        log.error('! failed to find %s', target_filepath)
        sys.exit()
    else:
        log.info('+ Downloaded: %s', target_filepath)


def build_osm_pbf_filepath(country_translated):
    """
    build download filepath to countries' OSM file
    replace / to have no problem with directories
    """
    # build path to downloaded file with translated geofabrik country
    map_file_path = os.path.join(
        USER_MAPS_DIR, f'{country_translated.replace("/", "_")}' + '-latest.osm.pbf')

    # return download filepath
    return map_file_path


def download_url_to_file(url, map_file_path):
    """
    download the content of a ULR to file
    """
    # set timeout to 30 minutes (per file)
    request_geofabrik = requests.get(
        url, allow_redirects=True, stream=True, timeout=1800)
    if request_geofabrik.status_code != 200:
        log.error('! failed download URL: %s', url)
        sys.exit()

    # write content to file
    write_to_file(map_file_path, request_geofabrik)


def download_tooling():
    """
    Windows
    - check for Windows tooling
    - download if Windows tooling is not available
    --> this is done to bring down the filesize of the python module

    macOS
    - check for mapwriter plugin and download if not existing
    """

    mapwriter_plugin_url = 'https://search.maven.org/remotecontent?filepath=org/mapsforge/mapsforge-map-writer/0.18.0/mapsforge-map-writer-0.18.0-jar-with-dependencies.jar'

    # Windows
    if platform.system() == "Windows":
        os.makedirs(USER_TOOLING_WIN_DIR, exist_ok=True)

        if not os.path.isfile(OSMOSIS_WIN_FILE_PATH):
            log.info('# Need to download Osmosis application for Windows')
            download_file(OSMOSIS_WIN_FILE_PATH,
                          'https://github.com/openstreetmap/osmosis/releases/download/0.48.3/osmosis-0.48.3.zip',
                          get_tooling_win_path('Osmosis', in_user_dir=True))

        if not os.path.isfile(get_tooling_win_path('osmfilter.exe', in_user_dir=True)):
            log.info('# Need to download osmfilter application for Windows')

            download_file(get_tooling_win_path('osmfilter.exe', in_user_dir=True),
                          'http://m.m.i24.cc/osmfilter.exe')

        mapwriter_plugin_path = os.path.join(USER_TOOLING_WIN_DIR,
                                             'Osmosis', 'lib', 'default', 'mapsforge-map-writer-0.18.0-jar-with-dependencies.jar')

    # Non-Windows
    else:
        mapwriter_plugin_path = os.path.join(
            str(USER_DIR), '.openstreetmap', 'osmosis', 'plugins', 'mapsforge-map-writer-0.18.0-jar-with-dependencies.jar')

    if not os.path.isfile(mapwriter_plugin_path):
        log.info('# Need to download Osmosis mapwriter plugin')
        # create plugins directory
        os.makedirs(os.path.dirname(mapwriter_plugin_path), exist_ok=True)
        download_file(mapwriter_plugin_path, mapwriter_plugin_url)

    # download geofabrik json as this will be needed always
    # because of the .json file replacement by geofabrik
    download_geofabrik_file_if_not_existing()


def download_geofabrik_file_if_not_existing():
    """
    check geofabrik file if not existing
    if the file does not exist, download geofabrik file
    """
    if not os.path.isfile(GEOFABRIK_PATH):
        log.info('# Need to download geofabrik file')
        download_file(GEOFABRIK_PATH,
                      'https://download.geofabrik.de/index-v1.json')


def get_latest_pypi_version():
    """
    get latest wahoomc version available on PyPI
    """
    try:
        response = requests.get(
            'https://pypi.org/pypi/wahoomc/json', timeout=1)
        return response.json()['info']['version']
    except (requests.ConnectionError, requests.Timeout):
        return None


def write_to_file(file_path, request):
    """
    write content of request into given file path
    """
    with open(file_path, mode='wb') as file_handle:
        for chunk in request.iter_content(chunk_size=1024*100):
            file_handle.write(chunk)


def unzip(source_filename, dest_dir):
    """
    unzip the given file into the given directory
    """
    with zipfile.ZipFile(source_filename, 'r') as zip_ref:
        zip_ref.extractall(dest_dir)


class Downloader:
    """
    This is the class to check and download maps / artifacts"
    """

    def __init__(self, max_days_old, force_download, border_countries=None):
        self.max_days_old = max_days_old
        self.force_download = force_download
        self.border_countries = border_countries

        # safety net if geofabrik file is not there
        # OsmData=>process_input_of_the_tool does it "correctly"
        download_geofabrik_file_if_not_existing()

        self.o_geofabrik_json = GeofabrikJson()

        self.need_to_dl = []

    def should_geofabrik_file_be_downloaded(self):
        """
        check geofabrik file if not existing or is not up-to-date

        # if geofabrik file needs to be downloaded, force-processing is set because there might be
        # a change in the geofabrik file
        """
        if self.check_file(GEOFABRIK_PATH) is True or \
                self.force_download is True:
            log.info('# Need to download geofabrik file')

            return True

        return False

    def download_geofabrik_file(self):
        """
        download geofabrik file
        """
        download_file(GEOFABRIK_PATH,
                      'https://download.geofabrik.de/index-v1.json')

        log.info('+ download geofabrik.json file: OK')

    def check_land_polygons_file(self):
        """
        check land_polygons file if not existing or are not up-to-date
        """

        if self.check_file(LAND_POLYGONS_PATH) is True or \
                self.force_download is True:
            log.info('+ Need to download land polygons file')
            self.need_to_dl.append('land_polygons')

    def download_files_if_needed(self):
        """
        check land_polygons and OSM map files if not existing or are not up-to-date
        and download if needed
        """
        if 'land_polygons' in self.need_to_dl:
            download_file(LAND_POLYGONS_PATH,
                          'https://osmdata.openstreetmap.de/download/land-polygons-split-4326.zip')

        # log.info('+ download land_polygons.shp file: OK')

        if 'osm_pbf' in self.need_to_dl:
            self.download_osm_pbf_file()

    def check_file(self, target_filepath):
        """
        check if given file is up-to-date
        """

        need_to_download = False
        logging_filename = target_filepath.rsplit('/', 1)[-1]

        log.info('-' * 80)
        log.info('# check %s file', logging_filename)

        # Check for expired file and delete it
        try:
            if self.should_file_be_downloaded(target_filepath):
                log.info('+ Deleting old %s file', logging_filename)
                os.remove(target_filepath)
                need_to_download = True

        except FileNotFoundError:
            need_to_download = True

        # if file does not exists --> download
        if not os.path.exists(target_filepath) or \
                not os.path.isfile(target_filepath):
            need_to_download = True

            log.info('+ %s file needs to be downloaded', logging_filename)

        return need_to_download

    def check_osm_pbf_file(self):
        """
        check if the relevant countries' OSM files are up-to-date
        """

        log.info('-' * 80)
        log.info('# check countries .osm.pbf files')

        # Check for expired maps and delete them
        log.info('+ Checking for old maps and remove them')

        for country in self.border_countries:
            # check for already existing .osm.pbf file
            map_file_path = glob.glob(
                f'{USER_MAPS_DIR}/{country}-latest.osm.pbf')
            if len(map_file_path) != 1:
                map_file_path = glob.glob(
                    f'{USER_MAPS_DIR}/**/{country}-latest.osm.pbf')

            # delete .osm.pbf file if out of date
            if len(map_file_path) == 1 and os.path.isfile(map_file_path[0]):
                if self.should_file_be_downloaded(map_file_path[0]):
                    log.info(
                        '+ mapfile for %s: deleted.', country)
                    os.remove(map_file_path[0])
                    self.need_to_dl.append('osm_pbf')
                else:
                    self.border_countries[country] = {
                        'map_file': map_file_path[0]}
                    log.info(
                        '+ mapfile for %s: up-to-date.', country)

            # mark country .osm.pbf file for download if there exists no file or it is no file
            map_file_path = self.border_countries[country].get('map_file')
            if map_file_path is None or (not os.path.isfile(map_file_path) or self.force_download):
                self.border_countries[country]['download'] = True
                self.need_to_dl.append('osm_pbf')

    def download_osm_pbf_file(self):
        """
        download countries' OSM files
        """
        for country, item in self.border_countries.items():
            try:
                if item['download'] is True:
                    # build path to downloaded file with translated geofabrik country
                    map_file_path = build_osm_pbf_filepath(country)
                    # fetch the geofabrik download url to countries' OSM file
                    url = self.o_geofabrik_json.get_geofabrik_url(country)
                    download_file(map_file_path, url)
                    self.border_countries[country] = {
                        'map_file': map_file_path}
            except KeyError:
                pass

    def should_file_be_downloaded(self, file_path):
        """
        check if given file should be downloaded
        - older that max_days old OR
        - force_download is set
        """

        chg_time = os.path.getmtime(file_path)

        return bool(older_than_x_days(chg_time, self.max_days_old) or self.force_download is True)
